#!/bin/sh
# AUTO-FAQ - automation for UseNet FAQ article maintenance
# Copyright (c) 1992, 1993, 1994 by Ian Kluft
# Copyright (c) 1995, 1999, 2002 by Paul W. Schleck
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
# USA.
#
# Questions or feedback regarding this program may be directed to:
#
# Paul W. Schleck
# auto-faq-maint@novia.net
# http://www.novia.net/~pschleck/auto-faq/
#
# P.O. Box 155
# Mount Rainier, MD  20712-0155 
# USA
# 
# RCS ID: $Id: auto-faq.SH,v 3.3 1997/07/26 14:30:53 pschleck Exp pschleck $
case $CONFIG in
'')
	if test -f config.sh; then TOP=.;
	elif test -f ../config.sh; then TOP=..;
	elif test -f ../../config.sh; then TOP=../..;
	elif test -f ../../../config.sh; then TOP=../../..;
	elif test -f ../../../../config.sh; then TOP=../../../..;
	else
		echo "Can't find config.sh."; exit 1
	fi
	. $TOP/config.sh
	;;
esac
rcs_revision=`echo '$Revision: 3.3 $' | sed 's/^[^ ]* \([^ \$]*\) .*/\1/'`
: This forces SH files to create target in same directory as SH file.
: This is so that make depend always knows where to find SH derivatives.
case "$0" in
*/*) cd `expr X$0 : 'X\(.*\)/'` ;;
esac
echo "Extracting auto-faq (with variable substitutions)"
: This section of the file will have variable substitutions done on it.
$spitshell >auto-faq <<!GROK!THIS!
#!$perlpath
#
# AUTO-FAQ - automation for UseNet FAQ article maintenance
# Copyright (c) 1992, 1993, 1994 by Ian Kluft
# Copyright (c) 1995, 1999, 2002 by Paul W. Schleck
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
# USA.
#
# Questions or feedback regarding this program may be directed to:
#
# Paul W. Schleck
# auto-faq-maint@novia.net
# http://www.novia.net/~pschleck/auto-faq/
#
# P.O. Box 155
# Mount Rainier, MD  20712-0155 
# USA
# 
# RCS ID: $Id: auto-faq.SH,v 3.3 1997/07/26 14:30:53 pschleck Exp pschleck $
#
#
# This script is intended for use from cron or at but will also run from
# the shell command line for manual FAQ posts.
#
# AUTO-FAQ uses the idea that there may be separate people with expertise
# in news and one or more FAQ topics.  This program should be run by the
# knowledgable news user in the group.  The FAQ file owners need to
# maintain their FAQs in files readable by the news user (with agreed-
# upon pathnames).
#
# environment variables used by this script:
# FAQ_TEST: if set, posts to test newsgroup instead of entire audience
# FAQ_NOSEND: if set, does not send mail or post news
# FAQ_DEBUG: enables variable tracing & on-the-fly status reports

# make a mail-id into a domain-qualified mail address
# if it already has an "@", "!", or "%", leave it as-is
sub domain_mail {
	local ( \$mail_id ) = @_;

	if ( \$mail_id =~ /[\@\%\!]/ ) {
		return ( \$mail_id );
	}
	return ( "\$mail_id\@\$config{'localdomain'}" );
}

# getdate - obtain parts of date/time from localtime
sub getdate
{
	local ( \$fielddesc, \$time ) = @_;
	if ( defined ( \$override_time )) {
		\$time = \$override_time;
	} elsif ( !defined( \$time )) {
		\$time = time;
	}
	local ( \$sec, \$min, \$hour, \$mday, \$mon, \$year, \$wday, \$isdst ) =
		localtime(\$time);

	# these numbers are straight out of struct tm - fix the month
	\$mon++;

	# use most of the System V date letters (aAbBdDehHImMprRSTwyYZ)
	# searches with regular expressions will catch first recognized char
	( \$fielddesc =~ /a/ ) && return (( "Sun", "Mon", "Tue", "Wed", "Thu", 
		"Fri", "Sat" )[\$wday] );
	( \$fielddesc =~ /A/ ) && return (( "Sunday", "Monday", "Tuesday",
		"Wednesday", "Thursday", "Friday", "Saturday" )[\$wday] );
	( \$fielddesc =~ /[bh]/ ) && return (( "Jan", "Feb", "Mar", "Apr",
		"May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
		)[\$mon-1] );
	( \$fielddesc =~ /B/ ) && return (( "January", "February", "March",
		"April", "May", "June", "July", "August", "September",
		"October", "November", "December" )[\$mon-1] );
	( \$fielddesc =~ /d/ ) && return ( sprintf ("%02d", \$mday ));
	( \$fielddesc =~ /D/ ) && return ( sprintf ("%02d/%02d/%02d", \$mon,
		\$mday, \$year % 100 ));
	( \$fielddesc =~ /e/ ) && return ( sprintf ("%2d", \$mday ));
	( \$fielddesc =~ /H/ ) && return ( sprintf ("%02d", \$hour ));
	( \$fielddesc =~ /I/ ) && return ( sprintf ("%02d", (\$hour==0 ? 12 :
		(\$hour<=12 ? \$hour : \$hour-12 ))));
	( \$fielddesc =~ /m/ ) && return ( sprintf ("%02d", \$mon ));
	( \$fielddesc =~ /M/ ) && return ( sprintf ("%02d", \$min ));
	( \$fielddesc =~ /p/ ) && return ( sprintf ("%s", (\$hour<12 ? "AM" :
		"PM" )));
	( \$fielddesc =~ /r/ ) && return ( sprintf ("%02d:%02d:%02d %s",
		(\$hour<12 ? \$hour : \$hour - 12), \$min, \$sec,
		(\$hour<12 ? "AM" :"PM" )));
	( \$fielddesc =~ /R/ ) && return ( sprintf ("%02d:%02d", \$hour,
		\$min ));
	( \$fielddesc =~ /S/ ) && return ( sprintf ("%02d", \$sec ));
	( \$fielddesc =~ /T/ ) && return ( sprintf ("%02d:%02d:%02d", \$hour,
		\$min, \$sec ));
	( \$fielddesc =~ /w/ ) && return ( sprintf ("%d", \$wday ));
	( \$fielddesc =~ /y/ ) && return ( sprintf ("%02d", \$year % 100 ));
	if ( \$fielddesc =~ /Y/ ) {
		# make sure this won't break at the turn of the century
		# we don't know that all versions of PERL will handle it alike
		\$turn_century=946684800;  # turn of the century in GMT
		local ( \$tc_foo, \$tc_foo, \$tc_foo, \$tc_foo, \$tc_foo, \$tc_year,
			\$tc_foo, \$tc_foo ) = gmtime(\$turn_century);
		if ( \$tc_year >= 100 && \$tc_year < 200 ) {
			# this version of PERL uses 100 for the year 2000
			return ( sprintf ("%04d", \$year + 1900 ));
		} else {
			# this version of PERL doesn't handle it as we expected
			# use the assumption that Unix handles years 1970-2038
			return ( sprintf ("%04d", (\$year%100) + ((\$year%100)
				< 70 ? 2000 : 1900 )));
		}
	}
	if ( \$fielddesc =~ /Z/ ) {
		\$ENV{"TZ"} =~ /^([^0-9]+)[0-9]([^0-9]+|)\$/;
		return ( \$isdst ? ( length(\$2)>0 ? \$2 : \$1 ) : \$1 );
	}

	# if we get here, the field descriptor string was not recognized
	print STDERR "\$0: getdate: bad date/time format '\$fielddesc'\n";
	return ( "" );
}

# evaluate a string for embedded attribute values
# resolves "@-macros" with limited recursion (i.e. limited = no loops)
sub eval_attr {
	local ( \$string ) = @_;
	local ( \$attr, \$val, %eval_attr );

	while ( \$string =~ /@([-A-Za-z0-9_]+)(\([-A-Za-z0-9_]+\)|)@/ ) {
		\$attr = \$1;
		\$arg = \$2; \$arg =~ s/^\((.*)\)/\$1/;
		\$faq_debug && print "eval_attr: attr=\$attr arg=\$arg\n";

		# handle functions
		if ( length(\$arg) > 0 ) {
			\$val="";
			if ( \$attr eq "date" ) {
				\$val = &getdate(\$arg);
			}
			if ( \$attr eq "posted_header" ) {
				\$val = \$saved_header{\$arg};
			}
                        if ( \$attr eq "seq_num" ) {
			        \$val = &get_seq_num(\$arg);
			}
			\$string =~ s/\@\$attr\(\$arg\)\@/\$val/;
		}
		# handle attributes
		if ( ! defined ( \$eval_attr{ \$attr })) {
			# if the attribute has not been evaluated before
			# then insert it
			\$eval_attr{ \$attr } = "x";
			\$val = \$config{ \$attr };
			\$string =~ s/\@\$attr\@/\$val/;
		}
	}
	return \$string;
}

# find the (optional) change control (SCCS/RCS/etc) file for the FAQ part
sub find_chgctrl_file {
	local ( \$part ) = @_;
	local ( @chgctrlpath, \$chgpath, \$chgctrlfile );

	if ( defined ( \$config{ 'chgctrl-path' } )) {
		\$chgpath = \$config{ 'chgctrl-path' };
	} else {
		if ( \$config{ 'chgctrl-type' } =~ /bsd|bsd-sccs/i
			|| length \$config{ 'chgctrl-type' } == 0 )
		{
			# default to BSD-style SCCS path
			\$chgpath = "SCCS/s.%s";
			if ( length(\$config{"chgctrl-format"}) == 0 ) {
				\$config{"chgctrl-format"}=":I: :Dm:/:Dd:/:Dy: :T:";
			}
		} elsif ( \$config{ 'chgctrl-type' } =~ /sccs|usl-sccs|att-sccs/i ) {
			# use USL-style SCCS path
			\$chgpath = "s.%s::SCCS/s.%s";
			if ( length(\$config{"chgctrl-format"}) == 0 ) {
				\$config{"chgctrl-format"}=":I: :Dm:/:Dd:/:Dy: :T:";
			}
		} elsif ( \$config{ 'chgctrl-type' } =~ /rcs/i ) {
			# use RCS paths
			\$chgpath = "%s,v::RCS/%s,v";
			if ( length(\$config{"chgctrl-format"}) == 0 ) {
				\$config{"chgctrl-format"}="-";
			}
		}
	}
	@chgctrlpath=split ( '::', \$chgpath );
	foreach \$path ( @chgctrlpath ) {
		if ( \$path =~ /^\// ) {
			\$chgctrlfile=sprintf ( \$path,
				"\$config{'articlebase'}.\$part" );
		} else {
			\$chgctrlfile=sprintf ( "\$config{'articledir'}/".\$path,
				"\$config{'articlebase'}.\$part" );
		}
		if ( -s \$chgctrlfile ) {
			\$faq_debug && print "chgctrlfile=\$chgctrlfile\n";
			return \$chgctrlfile;
		}
	}

	\$faq_debug && print "chgctrlfile not found ".
		"(not an error unless change control is in use)\n";
	return "";
}

# get change control revision/header info from the change control file
# this should only be called if the change control file was found
sub get_chgctrl_info {
	# parameters:
	# file: change control (SCCS/RCS) file
	# format: default format string
	#
	# NOTES:
	#    1) format will not be used if the chgctrl_cmd attribute is defined
	#    2) format and chgctrl_cmd may contain embedded config attributes
	#       in the @attribute-name@ ("@-macro") notation
	#       
	local ( \$file, \$format ) = @_;
	local ( \$tmpcmd, \$cmdline, \$string );

	if ( defined ( \$config{ 'chgctrl-cmd' })) {
		\$tmpcmd = \$config{ 'chgctrl-cmd' };
	} else {
		if ( \$config{ 'chgctrl-type' } =~ /bsd|bsd-sccs/i
			|| length \$config{ 'chgctrl-type' } == 0 )
		{
			# default to BSD-style SCCS command
			\$tmpcmd = 'sccs prs -d"'.\$format.'" %s';
		} elsif ( \$config{ 'chgctrl-type' } =~ /sccs|usl-sccs|att-sccs/i ) {
			# use USL-style SCCS command
			\$tmpcmd = 'prs -d"'.\$format.'" %s';
		} elsif ( \$config{ 'chgctrl-type' } =~ /rcs/i ) {
			# use RCS command
			\$tmpcmd = "rlog -b %s | awk ".
				'\'/^revision / {ver=\$2; getline; '.
				'tstamp=substr(\$0,6,20); print ver tstamp; '.
				"exit}'";
		}
	}
	\$cmdline = &eval_attr ( sprintf ( \$tmpcmd, \$file ));
	\$faq_debug && print "cmdline=\$cmdline\n";
	open ( CHGCTRLPIPE, "\$cmdline |" ) || return "";
	\$string=<CHGCTRLPIPE>;
	chop \$string;
	close CHGCTRLPIPE;
	\$faq_debug && print "chgctrl_string=\$string\n";
	return \$string;
}

# function to print a header
sub pr_header {
	local ( \$name, \$value ) = @_;

	\$saved_header{\$name}=\$value;
	print NEWSPIPE ( &eval_attr ( \$name.": ".\$value ))."\n";
}

#
# functions for printing each news header
# these can be called by name from a table
#
# The function name includes a prefix "hdr_" and the name of the header in
# all lower case.  (Dashes are also replaced with underscores in the subroutine
# name.)
#

sub hdr_from {
	if ( \$faq_test ) {
		&pr_header ( "From", &domain_mail(\$config{'newstech'}).
			" (\$config{'techname'})" );
	} else {
		if ( ! defined(\$config{'from'}) ||
			\$config{'from'} eq "newstech") {
			&pr_header ( "From",
				&domain_mail(\$config{'newstech'}).
				" (\$config{'techname'})" );
		} elsif ( \$config{'from'} eq "owner" ) {
			&pr_header ( "From",
				&domain_mail(\$config{'owner'}).
				" (\$config{'ownername'})" );
		} else {
			&pr_header ( "From", \$config{'from'} );
		}
	}
}

sub hdr_newsgroups {
	if ( \$faq_test ) {
		&pr_header ( "Newsgroups", "\$config{'testgroup'}" );
	} else {
		local ( \$extras ) = ( "" );
		if ( length ( \$config{'extragroups'} ) > 0 ) {
			\$extras= ",\$config{'extragroups'}";
		}
		&pr_header ( "Newsgroups", "\$config{'newsgroup'}\$extras" );
	}
}

sub hdr_subject {
	local ( \$trailer ) = ( "" );

	if ( \$total > 1 ) {
		\$trailer = " (Part \$part of \$total)";
	}
	if ( \$faq_test ) {
		\$trailer = \$trailer." [test]";
	}
	if ( length ( \$config{"title-\$part"}) > 0 ) {
		&pr_header ( "Subject", \$config{"title-\$part"});
	} elsif ( length ( \$config{'title'}) > 0 ) {
		&pr_header ( "Subject", \$config{'title'}.\$trailer );
	} else {
		&pr_header ( "Subject", "\$config{'newsgroup'} ".
			"Frequently Asked Questions".\$trailer );
	}
}

sub hdr_message_id {
	# same in test and post modes
	if ( length ( \$config{'message-id'}) > 0 ) {
		&pr_header ( "Message-ID", "<\$config{'message-id'}" );
	} elsif ( length ( \$config{'faqid'}) > 0 ) {
		&pr_header ( "Message-ID", "<\$config{'faqid'}-".
			"\$part-\$seqnum\@\$config{'localdomain'}>" );
	}
}

sub hdr_references {
	# same in test and post modes
	\$references="";
	if (( length ( \$config{'faqid'}) > 0 ) && ( \$part > 1 )) {
		\$references = "<\$config{'faqid'}-".
			"1-\$seqnum\@\$config{'localdomain'}> ";
	}
	if (( length ( \$config{'refid'}) > 0 ) &&
		( length ( \$config{'refseqfile'}) > 0 ))
	{
		if ( open ( REFFILE, "\$config{'refseqfile'}" )) {
			defined ( REFFILE ) &&
			( \$refnum = <REFFILE> ) && chop \$refnum &&
				( \$references = "\$references".
				"<\$config{'refid'}-".
				"1-\$refnum\@\$config{'localdomain'}> " );
			close REFFILE;
		}
	}
	if ( length ( \$config{"Reference-message-id"}) > 0 ) {
		\$references = "\$references \@Reference-message-id\@";
	}
	if ( length ( \$references ) > 0 ) {
		&pr_header ( "References", "\$references" );
	}
}

sub hdr_reply_to {
	if ( \$faq_test ) {
		&pr_header ( "Reply-To", &domain_mail(\$config{'newstech'}).
			" (\$config{'techname'})" );
	} else {
		&pr_header ( "Reply-To", &domain_mail(\$config{'owner'}).
			" (\$config{'ownername'})" );
	}
}

sub hdr_followup_to {
	if ( length ( \$config{'followup'} ) > 0 ) {
		&pr_header ( "Followup-To", "\$config{'followup'}" );
	} else {
		if ( \$faq_test ) {
			&pr_header ( "Followup-To", "\$config{'testgroup'}" );
		} else {
			&pr_header ( "Followup-To", "\$config{'newsgroup'}" );
		}
	}
}

sub hdr_distribution {
	if ( \$faq_test ) {
		&pr_header ( "Distribution", "\$config{'testdistrib'}" );
	} else {
		&pr_header ( "Distribution", "\$config{'distribution'}" );
	}
}

sub hdr_organization {
	&pr_header ( "Organization", "\$config{'organization'}" );
}

sub hdr_expires {
	# post mode will be in primary hdr, test mode in test-tertiary hdr
	&pr_header ( "Expires", "\$expdate" );
}

sub hdr_approved {
	# post mode will be in primary hdr, test mode in test-tertiary hdr
	if ( length ( \$config{'approval'} ) > 0 ) {
		&pr_header ( "Approved", "\$config{'approval'}" );
	}
}

sub hdr_supersedes {
	# post mode will be in primary hdr, test mode in test-tertiary hdr
	if ( length(\$config{'Saved-message-id'}) > 0 ) {
		&pr_header ( "Supersedes",
			"\$config{'Saved-message-id'}" );
	} elsif ( length(\$supnum) > 0 && length ( \$config{'faqid'}) > 0 ) {
		&pr_header ( "Supersedes", "<\$config{'faqid'}-".
			"\$part-\$supnum\@\$config{'localdomain'}>" );
	}
}

sub hdr_posted_by {

        # Different versions of Perl might have round-off error in
        # evaluating \$] (the Perl version) in a numeric context
        \$perl_version = sprintf("%5.3f", \$]);
	if ( \$faq_test ) {
#		&pr_header ( "Posted-By", "auto-faq \$rcs_revision (test mode)" );
		&pr_header ( "Posted-By", "auto-faq 3.3.1 (Perl \$perl_version) (test mode)" );
	} else {
#		&pr_header ( "Posted-By", "auto-faq \$rcs_revision" );
		&pr_header ( "Posted-By", "auto-faq 3.3.1 (Perl \$perl_version)" );
	}
}

sub hdr_archive_name {
	# same in test and post modes
	local ( \$trailer ) = ( "" );

	if ( \$total > 1 ) {
		\$trailer = "/part\$part";
	}
	if ( length ( \$config{'archive'}) > 0 ) {
		&pr_header ( "Archive-name", "\$config{'archive'}".\$trailer );
	} elsif ( length ( \$config{'faqid'}) > 0 ) {
		&pr_header ( "Archive-name", "\$config{'faqid'}".\$trailer );
	}
}

sub hdr_revision {
	# same in test and post modes
	if ( length ( \$chgctrlfile ) > 0 ) {
		if ( length ( \$config{'chgctrl-format'}) > 0 ) {
			\$chgctrl_string=&get_chgctrl_info(
				\$chgctrlfile, "\@chgctrl-format\@" );
			if ( length ( \$chgctrl_string ) > 0 ) {
				&pr_header ( "Revision",
					"\$chgctrl_string" );
			}
		}
	}
}

sub hdr_changes {
	# same in test and post modes
	if ( length ( \$chgctrlfile ) > 0 ) {
		if ( length ( \$config{'change-format'}) > 0 ) {
			\$chgctrl_string=&get_chgctrl_info(
				\$chgctrlfile, "\@change-format\@" );
			if ( length ( \$chgctrl_string ) > 0 ) {
				&pr_header ( "Changes",
					"\$chgctrl_string" );
			}
		}
	}
}

sub call_hdr_func {
	local ( \$header ) = @_;
	\$header =~ tr/-A-Z/_a-z/;

	# use PERL's name referencing to get functions from symbol table
	local ( *local_func, *system_func ) =
		( "local_hdr_\$header", "hdr_\$header" );

	# if a local-site-override function exists, use it instead
	if ( defined &local_func ) {
		\$faq_debug && print "calling local_hdr_\$header()\n";
		&local_func;  # call function by name
	} elsif ( defined &system_func ) {
		\$faq_debug && print "calling hdr_\$header()\n";
		&system_func;  # call function by name
	} else {
		\$faq_debug && print "no header function for \$header\n";
		print STDERR "auto-faq: no internal or local function exists ".	
			"for the \$header header\n";
	}
}

# faq_article - function to print article headers and contents
#
# variables expected to be set before entry:
#    NEWSPIPE	- file descriptor used for output
#    \$expdate	- article expiration date
#    \$seqnum	- sequence number for this article
#    \$keyword	- faq keyword/name (matches cmd-line to items in config file)
#		  the keyword should be related to the newsgroup name/topic
#
# configuration parameters in the "config" associative array
#    testgroup	- newsgroup used for test postings
#    testdistrib - distribution used for test postings
#    extragroups - newsgroup name(s) for cross-posting (comma delimited)
#    owner	- owner of the FAQ file
#    ownername	- full name of the owner of the FAQ file
#    newstech	- technical lead for this FAQ posting (may be same as owner)
#    techname	- fullname of the newstech
#    articledir - directory path name for FAQ files
#    articlebase - base name for FAQ files (appends .1, .2, ...)
#    organization - for Organization header line
#    distribution - for Distribution header line
#    faqid	- FAQ id used for message ID & archive name
#    localdomain - local internet domain name (for message id & return mail)
#    approval	- news.answers approval (Approval header omitted if blank)
#    timezone   - used to set the TZ environment variable (default is GMT)
#
# variables that are used but not required to be set before entry:
#    \$supnum	- sequence number of article this one supercedes
#

# print the FAQ article including configured header fields
sub faq_article
{
	local ( \$part, \$total ) = @_;
	local ( \$chgctrl_string, \$statusfile );

	# make part number available in configuration
	\$config{"part"}=\$part;

	# initialize filenames
	\$articlefile="\$config{'articledir'}/\$config{'articlebase'}.\$part";
	\$chgctrlfile=&find_chgctrl_file ( \$part );

	# load per-part status file if it exists
	if ( length ( \$config{"statusbase"}) > 0 ) {
		# derive status file path - also used later for writing
		\$statusfile=&eval_attr("\@articledir\@/\@statusbase\@.\@part\@");

		# same format as a config file
		# all attributes prefixed with "Saved-"
		if ( -r \$statusfile && -s \$statusfile ) {
			if (( \$cfgstatus = &getconfig( \$statusfile, "",
				"Saved-" )) <= 0 )
			{
				# fatal if status file fails
				( \$cfgstatus == -1 ) &&
					die "\$0: could not open status ".
						"file \$statusfile\n";
				die "\$0: error reading status file ".
					"\$statusfile\n";
			}
		}
	}

	# perform posting for test or post mode
	if ( \$faq_test ) {
		# test mode
		foreach \$header ( @test_primary ) {
			&call_hdr_func ( \$header );
		}
		for \$header ( @headers ) {
			\$header =~ /([^\s]*): /;
			&pr_header ( "\$1", "\$'" );
		}
		print NEWSPIPE "\n";
		foreach \$header ( @test_secondary ) {
			&call_hdr_func ( \$header );
		}
		for \$auxheader ( @auxheaders ) {
			\$auxheader =~ /([^\s]*): /;
			&pr_header ( "\$1", "\$'" );
		}
		print NEWSPIPE "\n";
		print NEWSPIPE "The following headers would also have been ".
			"included if this had not been a test:\n";
		foreach \$header ( @test_tertiary ) {
			&call_hdr_func ( \$header );
		}
		print NEWSPIPE "\n";
	} else {
		# the real thing
		foreach \$header ( @post_primary ) {
			&call_hdr_func ( \$header );
		}
		for \$header ( @headers ) {
			\$header =~ /([^\s]*): /;
			&pr_header ( "\$1", "\$'" );
		}
		print NEWSPIPE "\n";
		foreach \$header ( @post_secondary ) {
			&call_hdr_func ( \$header );
		}
		for \$auxheader ( @auxheaders ) {
			\$auxheader =~ /([^\s]*): /;
			&pr_header ( "\$1", "\$'" );
		}
		print NEWSPIPE "\n";
	}

	# copy article text to posted article
	open(ARTICLEFILE, \$articlefile);
	while (<ARTICLEFILE>) {
		print NEWSPIPE \$_;
	}
	close(ARTICLEFILE);

	# write out the per-part status file to save info for next article
	if ( ! \$faq_test && ! \$faq_nosend && length(\$statusfile) > 0 ) {
		open ( STATUSFILE, ">\$statusfile" ) ||
			die "could not open \$statusfile for writing: \$!\n";
		if ( defined \$config{"savestatus"}) {
			@savestatus=split(/::/, \$config{"savestatus"});
		} else {
			@savestatus=(
				"message-id=\@posted_header(Message-ID)\@",
				"date=\@date(a)\@ \@date(d)\@ \@date(b)\@ ".
					"\@date(T)\@ \@date(Z)\@"
				);
		}
		foreach \$attr ( @savestatus ) {
			\$attr =~ s/([^=]+)=(.*)/\$1=\"\$2\"/; # add missing quotes
			print STATUSFILE (&eval_attr(\$attr))."\n";
		}
		close STATUSFILE;
	}
}

# read configuration file
sub getconfig
{
	local ( \$configfile, \$keyword, \$prefix, \$level, \$found,
		\$readgroup, \$global ) = @_;
	local ( \$tmp_header, \$tmp_h2, \$separator );

	# set up local variables
	if ( ! defined \$level ) {
		# called from subroutine other than getconfig
		\$level=0;
		\$found=0;
		\$global=1; # accept global attrs until 1st "=== keyword" line
		\$readgroup=""; # name of group of attributes being scanned
			# null readgroup is redundant w/ global for readbility
	} else {
		# recursive call from getconfig (for an include file)
		\$level++;
	}

	# open the config file
	if ( ! open(CONFIGFILE, \$configfile) ) {
		print STDERR "auto-faq: could not open config file ".
			"\$configfile\n";
		if ( \$level == 0 ) {
			return -1;
		} else {
			# failure of a subsidiary config file is not
			# fatal but should be considered serious

			# return from recursive call back to getconfig
			return ( \$found, \$readgroup, \$global );
		}
	}
	\$faq_debug && print "reading config: \$configfile (level=\$level)\n";

	# process each line in the config file
	while (<CONFIGFILE>) {
		/^#/ && next;		# skip comments
		/^[ \t\n]*\$/ && next;	# skip blank lines
		if ( /^===\s*(\S+)\s*/ ) {
			\$global=0;
			\$readgroup = \$1;
			\$faq_debug && print "readgroup=\$readgroup\n";
			next;
		}
		if ( /^=include (.*)/ ) {
			\$faq_debug && print "including config: ".
				"\$configfile (level=\$level)\n";
			( \$found, \$readgroup, \$global ) =
				&getconfig( \$1, \$keyword,
				\$level, \$prefix, \$found,
				\$readgroup, \$global );
		}
		( ! \$global && \$readgroup ne \$keyword && \$found == 1 ) && last;
		( ! \$global && \$readgroup ne \$keyword ) && next;
		! \$global && ( \$found=1 );
		if ( /^([^ =]+)=\"([^"]*)\"/ ) {
			( \$faq_debug ) && print "config: \$1=\"\$2\"\n";
			\$config{\$prefix.\$1}=\$2;
		} elsif ( /^([^ \t:]+)(:+) ([^\n]*)/ ) {
			\$tmp_header="\$1: \$3"; \$tmp_h2=""; \$separator=\$2;
			while (( \$tmp_header =~ /\\\\\$/ ) &&
				( \$tmp_h2 = <CONFIGFILE>))
			{
				chop \$tmp_h2;
				( \$faq_debug ) && print "buffer: \$tmp_h2\n";
				\$tmp_header =~ s/\\\\\$//;
				\$tmp_header = "\$tmp_header\n\$tmp_h2";
				# note: if a \-terminated line is followed by
				# an EOF, the \ will remain so it will catch
				# the news guru's attention
			}
			( \$faq_debug ) && print "header: \$tmp_header\n";
			if ( length( \$separator ) == 1 ) {
				push ( @headers, \$tmp_header );
			} elsif ( length( \$separator ) == 2 ) {
				push ( @auxheaders, \$tmp_header );
			}
		}
	}
	close CONFIGFILE;

	# return value depends on whether this is a recursive call or not
	# (additional info returned from recursive calls)
	if ( \$level == 0 ) {
		# return to calling subroutine
		if ( length ( \$keyword ) > 0 )
		{
			# if a keyword was specified, return success/failure
			return ( \$found );
		} else {
			# weren't looking for any keyword, always success
			return ( 1 );
		}
	} else {
		# return from recursive call back to getconfig
		return ( \$found, \$readgroup, \$global );
	}
}

sub postfaq {
	local ( \$keyword ) = @_;
	local ( \$maxfaq, \$expformat, \$cfgstatus, \$refstatus );

	# obtain configuration parameters
	\$faq_debug && print "keyword: \$keyword\n";
	\$cfgfilename = "faq-config";
	if (( \$cfgstatus =
		&getconfig( \$cfgfilename, \$keyword, "" )) <= 0 )
	{
		# failure to process the top-level config file is fatal
		( \$cfgstatus == 0 ) &&
			die ( "\$0: keyword \$keyword not found in config ".
				"file\n" );
		( \$cfgstatus == -1 ) &&
			die ( "\$0: could not open config file ".
				"\$cfgfilename\n" );
		die ( "\$0: error reading config file \$cfgfilename\n" );
	}

	# load referenced-article status, if requested
	if ( length ( \$config{"refstatusfile"}) > 0 ) {
		# derive reference status file
		\$refstatusfile=&eval_attr("\@articledir\@/\@refstatusfile\@.\@part\@");

		# same format as a config file
		# all attributes prefixed with "Reference-"
		if ( -r \$refstatusfile && -s \$refstatusfile ) {
			if (( \$refstatus = &getconfig( \$refstatusfile, "",
				"Reference-" )) <= 0 )
			{
				# fatal if referenced status file fails
				( \$refstatus == -1 ) &&
					die "\$0: could not open status ".
						"file \$refstatusfile\n";
				die "\$0: error reading status file ".
					"\$refstatusfile\n";
			}
		}
	}


	# set the time zone
	if ( defined ( \$config{'timezone'})) {
		\$ENV{"TZ"}=\$config{'timezone'};
	} else {
		\$ENV{"TZ"}="GMT";
	}

	# double-check that we're not going to overwrite the text files
	if ( \$config{"articlebase"} eq \$config{"statusbase"} ) {
		die "\$0: statusbase cannot be same as articlebase\n".
			"   would overwrite text file\n";
	}

	# make sure FAQ owner's part is there
	\$maxfaq=1;
	while ( -f ( \$afile="\$config{'articledir'}/\$config{'articlebase'}.".
		"\$maxfaq" ))
	{
		if ( ! -r \$afile ) {
			# if any aren't readable, send mail
			if ( \$faq_nosend ) {
				\$mailpipe= "| { echo ".
					"'--- would have sent mail ---'; ".
					"$cat; echo '--- end of mail ---'; ".
					"} >> \$logfile";
			} else {
				\$mailpipe="| $mailer ".
					&domain_mail(\$config{"owner"});
			}
			open(MAILPIPE, \$mailpipe)
				|| die "could not send mail to ".
				&domain_mail(\$config{'owner'});

			print(MAILPIPE "Subject: \$keyword FAQ posting failed\n");
			print MAILPIPE "\n";
			print MAILPIPE "[Sent by cron job]\n";
			print MAILPIPE "\$afile is not readable\n";
			print MAILPIPE "please post the \$keyword ".
				"FAQ manually\n";
			close(MAILPIPE);

			die "\$keyword FAQ posting failed: \$afile unreadable";
		}
		\$maxfaq++;
	}
	\$maxfaq--;
	if ( \$maxfaq <= 0 ) {
		# couldn't find any FAQ files, better send mail
		if ( \$faq_nosend ) {
			\$mailpipe="| { echo '--- would have sent mail ---'; ".
				"$cat; echo '--- end of mail ---'; ".
				"} >> \$logfile";
		} else {
			\$mailpipe="| $mailer ".&domain_mail(\$config{"owner"});
		}
		open(MAILPIPE, \$mailpipe)
			|| die "could not send mail to \$config{'owner'}";

		# if any aren't readable, send mail
		print MAILPIPE "Subject: \$keyword FAQ posting failed\n";
		print MAILPIPE "\n";
		print MAILPIPE "[Sent by cron job]\n";
		print MAILPIPE "The \$keyword FAQ file(s) in ".
			"\$config{'articledir'}/\$config{'articlebase'}.* ".
			"are either\n";
		print MAILPIPE "missing or unreadable.  Please ".
			"post the FAQ manually.\n";
		die "\$keyword FAQ posting failed: ".
			"\$config{'articledir'}/\$config{'articlebase'}.\* ".
			"not found";
	}

	\$faq_debug && print "maxfaq=\$maxfaq\n";

	# set up default header lists
	@test_primary = ( "From", "Newsgroups", "Subject", "Message-ID",
		"References", "Reply-To", "Followup-To", "Distribution",
		"Organization" );
	@test_secondary = ( "Posted-By", "Archive-Name", "Revision",
		"Changes" );
	@test_tertiary = ( "Approved", "Expires", "Supersedes" );
	@post_primary = ( "From", "Newsgroups", "Subject", "Message-ID",
		"References", "Reply-To", "Followup-To", "Distribution",
		"Organization", "Approved", "Expires", "Supersedes" );
	@post_secondary = ( "Posted-By", "Archive-Name", "Revision",
		"Changes" );

	# allow user to change header list contents or order
	if ( length ( \$config{'test-pri-hdrs'}) > 0 ) {
		@test_primary = split ( /:/, \$config{'test-pri-hdrs'});
	}
	if ( length ( \$config{'test-2nd-hdrs'}) > 0 ) {
		@test_secondary = split ( /:/, \$config{'test-2nd-hdrs'});
	}
	if ( length ( \$config{'test-3rd-hdrs'}) > 0 ) {
		@test_tertiary = split ( /:/, \$config{'test-3rd-hdrs'});
	}
	if ( length ( \$config{'post-pri-hdrs'}) > 0 ) {
		@post_primary = split ( /:/, \$config{'post-pri-hdrs'});
	}
	if ( length ( \$config{'post-2nd-hdrs'}) > 0 ) {
		@post_secondary = split ( /:/, \$config{'post-2nd-hdrs'});
	}

	# load external functions if requested
	#
	#    NOTE: this feature will not work if auto-faq is run with
	#    taintperl (or any setuid PERL) because it tries to execute
	#    PERL code read from a file.  The file input is considered
	#    "tainted" and, therefore, incompatible with the setuid PERLs.
	#    Solution: if you must use setuid to any uid, you cannot use
	#    user-loadable header functions - you're stuck with the built-ins.
	#
	if ( length ( \$config{'local-func-src'}) > 0 ) {
		local ( @source );

		if ( open ( LOCALHDR, \$config{'local-func-src'})) {
			# gulp the file
			@source = <LOCALHDR>;

			# we're done with the file now - we have it in memory
			close LOCALHDR;

			# interpret the code, report any errors
			\$faq_debug && print "interpreting ".
				"\$config{'local-func-src'}\n";
			eval "\@source";
			if ( length ( \$@ ) > 0 ) {
				print STDERR "auto-faq: errors found in ".
					\$config{'local-func-src'}."\n";
				print STDERR "\$@";
			}
		} else {
			print STDERR "auto-faq: could not open header ".
				"sources ".\$config{'local-func-src'}."\n";
		}
	}

	#
	# ready...         (prepare to fire off the articles)
	#

	# calculate expiration date
	if ( length(\$config{"expires-format"}) == 0 ) {
		# default is taken from recommendation in RFC 1036
		\$expformat="\@date(a)\@ \@date(d)\@ \@date(b)\@ \@date(Y)\@ \@date(T)\@ \@date(Z)\@";
	} else {
		\$expformat=\$config{"expires-format"};
	}
	\$override_time = time + int(\$config{'interval'})*24*60*60; # add n days
	\$expdate=&eval_attr(\$expformat);
	undef ( \$override_time );

	# calculate sequence number
	# try sequence file, will be able to use per-part status file later too
	\$seqnum=time;
	\$seqfile=\$config{"seqfile"};
	if ( length ( \$seqfile ) > 0 ) {
		# use a sequence file
		if ( -r \$seqfile && -s \$seqfile ) {
			if ( open ( SEQFILE, \$seqfile ) ) {
				\$supnum = <SEQFILE>;
				chop \$supnum;
				close SEQFILE;
			}
			\$faq_debug && -t 1 && print "supnum=\$supnum\n";
		}
	}

	#
	# aim... FIRE!     (post them)
	#
	for ( \$counter=1; \$counter <= \$maxfaq; \$counter++ ) {
		if ( \$faq_nosend ) {
			\$newspipe="| ".
				"{ echo '--- would have posted news ---'; ".
				"$cat; } >> \$logfile";
		} else {
			if ( length ( \$config{ 'newscmd' } ) > 0 ) {
				\$newspipe="| \$config{ 'newscmd' }";
			} else {
				\$newspipe="| $newslibexp/inews -h";
			}

			# strip off any leading "|" and trailing "-h" in order to test the path
			\$newspipe =~ /^\|?\s*(\S+)\s*(\-h|)\s*\$/;
			\$base_newscmd = \$1;
			\$faq_debug && -t 1 && print "base_newscmd=\$base_newscmd\n";

			# test path because if the path is not valid, a cryptic
			# "broken pipe" error message will be displayed
			# (open to a pipe only fails if fork fails)
			if ( -d \$base_newscmd ) {
				die "faq-config error, newscmd: \$base_newscmd is a directory, not an executable file\n";
			}
			if ( ! -f \$base_newscmd ) {
				die "faq-config error, newscmd: \$base_newscmd not found, no such file or directory\n";
			}
			if ( ! -x \$base_newscmd ) {
				die "faq-config error, newscmd: \$base_newscmd is not executable\n";
			}

		}
		open (NEWSPIPE, \$newspipe) ||
			die "Could not run news command for \$keyword FAQ\n";
		
		&faq_article ( \$counter, \$maxfaq );

		# check for exit code of news command, and give appropriate diagnostic message
		# if necessary, then exit program to avoid writing out sequence information
		# (as an error indicates that the article was not posted)
		close NEWSPIPE || die "\nError closing NEWSPIPE: ".(\$! || "exit status \$?")."\n";

	}

	# write out the sequence file to save info for next article
	if ( ! \$faq_test && ! \$faq_nosend && length(\$seqfile) > 0 ) {
		open ( SEQFILE, ">\$seqfile" ) ||
			die "could not open \$seqfile for writing: \$!\n";
		print SEQFILE "\$seqnum\n";
		close SEQFILE;
	}
}

sub get_seq_num {

# Generate a one-up sequence number using a sequence number file in an
# idempotent way (i.e., only one sequence number per \$keyword),
# intended to be used as a @seq_num@ macro.  The argument determines
# the starting number of the offset, such as for integrating auto-faq
# into an FAQ posting scheme where you are already on sequence number
# 975, or whether or not you want a 0 or 1 offset sequence number.
# Having an argument is also a convenient way of meeting the
# requirement within auto-faq that macros that are the returned values
# of functions be of the form "@name(arg)@"

  local(\$offset) = shift @_;
  local(\$seq_num, \$return_seq_num);

  \$seq_numfile = "\$config{'articledir'}/\$config{'seq_numbase'}";
  # check for cached value, else get from sequence file
  if (\$saved_seq_num{\$keyword}) {
  	\$seq_num = \$saved_seq_num{\$keyword};
       	\$faq_debug && -t 1 && print "(cached) seq_num=\$seq_num\n";
  } else {
	 if ( -r \$seq_numfile && -s \$seq_numfile ) {
	       if ( open ( SEQ_NUMFILE, \$seq_numfile ) ) {
			\$seq_num = <SEQ_NUMFILE>;
	                chop \$seq_num;
		        \$seq_num = \$seq_num + 1;
			close SEQ_NUMFILE;
		}
		\$faq_debug && -t 1 && print "seq_num=\$seq_num\n";
	 } else {
		# file not found, start with 1	
		\$seq_num = 1;
	 }
	 if (! \$faq_test && ! \$faq_nosend && length(\$seq_numfile) > 0 ) {
		open (SEQ_NUMFILE, ">\$seq_numfile" ) ||
			die "could not open \$seq_numfile for writing: \$!\n";
		print SEQ_NUMFILE "\$seq_num\n";
		close SEQ_NUMFILE;
	 }
	 # cache for possible multiple parts of \$keyword FAQ
	 \$saved_seq_num{\$keyword} = \$seq_num;
  }

  # \$return_seq_num is seq_num with \$offset applied
  \$return_seq_num = \$seq_num - (1 - \$offset);

  return \$return_seq_num;
}

# set up a minimal environment since this will be started from cron
# TZ will be set from the config file
\$ENV{"HOME"}=\`$basicshell -c pwd\`;
\$ENV{"HOME"} =~ s/\s//g;
\$ENV{"PATH"}=\$ENV{"PATH"}.":/usr/local/bin";
\$ENV{"DOTDIR"}=\`$basicshell -c pwd\`;
\$ENV{"DOTDIR"} =~ s/\s//g;
umask(0022);

# check for test environment flags
\$faq_test = length (\$ENV{"FAQ_TEST"}) > 0;
\$faq_nosend = length (\$ENV{"FAQ_NOSEND"}) > 0;
\$faq_debug = length (\$ENV{"FAQ_DEBUG"}) > 0;

# check command line
if ( \$#ARGV < 0 ) {
	die ( "usage: \$0 faq-name [...]\n" );
}

# set up debugging & testing options
\$faq_debug && ( \$| = 1 );	# make sure pipes are flushed in debug mode
if ( \$faq_nosend ) {
	\$logfile = "faq.out";
	unlink ( \$logfile );
}

# post one FAQ per word on the command-line
foreach \$keyword (@ARGV) {
	undef %config;
	undef @headers;
	undef @auxheaders;
	eval "&postfaq ( '".\$keyword."' )" || print "\$@";
}

!GROK!THIS!

: In the following dollars and backticks do not need the extra backslash.
$spitshell >>auto-faq <<'!NO!SUBS!'
!NO!SUBS!
chmod 755 auto-faq
$eunicefix auto-faq
